from __future__ import annotations

from functools import wraps

import plotly.express as px
from plotlab import Figure

# from .panel import Panel

# TODO: Set plotting configs and add kwargs to functions
# TODO: Check if kwargs would overwrite fig.add_trace if same params are used


class PanelFigure(Figure):
    """
    PanelFigure class.
    """

    def __init__(self):
        # TODO: Add dynamic color changing once new traces are added
        # TODO: Add candlestick plot
        # TODO: Add trace names as panel cols
        super().__init__()
        self.colors = px.colors.qualitative.Plotly
        self.color_index = 0

    def add_annotation(self, panel, color: str = "gray", opacity: float = 1.0) -> None:
        """
        Plot vertical lines showing train, val, and test periods.

        Args:
            panel (``Panel``): Panel to split.
            color (``str``): Color of the sets.
            opacity (``float``): Opacity of the sets.
        """
        # BUG: Seems to break if using "ggplot2"
        # ! Won't take effect until next trace is added (no axis was added)

        ymax = panel.max().max() if panel.train_size else 0

        if hasattr(panel, "train_size") and panel.train_size:
            xtrain_min = panel.train.index.min()
            self.fig.add_vline(
                x=xtrain_min, line_dash="dot", line_color=color, opacity=opacity
            )
            self.fig.add_annotation(
                x=xtrain_min, y=ymax, text="Train", showarrow=False, xshift=20
            )

        if hasattr(panel, "val_size") and panel.val_size:
            xval_min = panel.val.index.min()
            self.fig.add_vline(
                x=xval_min, line_dash="dot", line_color=color, opacity=opacity
            )
            self.fig.add_annotation(
                x=xval_min, y=ymax, text="Validation", showarrow=False, xshift=35
            )

        if hasattr(panel, "test_size") and panel.test_size:
            xtest_min = panel.test.index.min()
            self.fig.add_vline(
                x=xtest_min, line_dash="dot", line_color=color, opacity=opacity
            )
            self.fig.add_annotation(
                x=xtest_min, y=ymax, text="Test", showarrow=False, xshift=18
            )

    # Add decorator for instance check and for loop
    def _iterator(func):
        @wraps(func)
        def inner(self, *args, **kwargs):

            args = list(args)
            df = args.pop(0)

            for col in df.columns:
                kwargs["color"] = self.colors[self.color_index]
                self.color_index = (self.color_index + 1) % len(self.colors)
                if col != "frame":
                    func(self, df[col], *tuple(args), **kwargs)

        return inner

    @_iterator
    def add_line(self, col: str, *args, **kwargs) -> None:
        """
        Add a line to the figure.

        Args:
            col (``str``): Column to plot
        """
        self.line(col, *args, **kwargs)

    @_iterator
    def add_area(self, col: str, *args, **kwargs) -> None:
        """
        Add an area to the figure.

        Args:
            col (``str``): Column to plot
        """
        self.area(col, *args, **kwargs)

    @_iterator
    def add_bar(self, col: str, *args, **kwargs) -> None:
        """
        Add a bar to the figure.

        Args:
            col (``str``): Column to plot
        """
        self.bar(col, *args, **kwargs)

    @_iterator
    def add_scatter(self, col: str, *args, **kwargs) -> None:
        """
        Add a scatter to the figure.

        Args:
            col (``str``): Column to plot.
        """
        self.scatter(col, *args, **kwargs)

    @_iterator
    def add_dotline(self, col: str, *args, **kwargs) -> None:
        """
        Add a dotline to the figure.

        Args:
            col (``str``): Column to plot.
        """
        self.dotline(col, *args, **kwargs)


def plot(
    panel, use_timestep: bool = False, add_annotation: bool = False, **kwargs
) -> PanelFigure:
    """
    Plot panel.

    Args:
        panel (``Panel``): Panel object.
        use_timestep (``bool``): Use timestep instead of id.
        add_annotation (``bool``): If True, plot vertical lines showing train, val, and test periods.

    Returns:
        ``Plot``: Plotted data
    """
    fig = PanelFigure()

    for col in panel.columns:
        if panel[col].dtype == bool:
            panel[col] = panel[col].astype(int)

    if use_timestep:
        panel = panel.droplevel(0, axis=0) if panel.index.nlevels > 1 else panel
    else:
        panel = panel.droplevel(1, axis=0) if panel.index.nlevels > 1 else panel

    fig.add_line(panel, **kwargs)

    if add_annotation and isinstance(panel.train_size, int):
        fig.add_annotation(panel)

    return fig()



from pathlib import Path

import pkg_resources
import toml

# from wavy.models import *
from wavy.panel import *


def get_version():
    try:
        return pkg_resources.get_distribution("wavyts").version
    except Exception:
        path = Path(__file__).resolve().parents[2] / "pyproject.toml"
        pyproject = toml.loads(open(str(path)).read())
        return pyproject["tool"]["poetry"]["version"]


__version__ = get_version()



from __future__ import annotations

# import contextlib
import random
import uuid
import warnings
from itertools import chain

import numpy as np
import pandas as pd
from matplotlib.pyplot import figure
from pandas.core.groupby import DataFrameGroupBy

warnings.simplefilter(action="ignore", category=FutureWarning)

from wavy.plot import plot
from wavy.validations import _validate_sample_panel, _validate_training_split


# Define column to y
def create_panels(
    df: pd.DataFrame, lookback: int, horizon: int, gap: int = 0
) -> tuple[Panel, Panel]:
    """
    Create panels from a dataframe.

    Args:
        df (``pd.DataFrame``): Dataframe
        lookback (``int``): Lookback size
        horizon (``int``): Horizon size
        gap (``int``): Gap size

    Returns:
        ``tuple[Panel, Panel]``: Tuple of panels
    """

    indices = df.index

    # Sort by index
    df = df.sort_index(ascending=True)

    if not all(df.index == indices):
        warnings.warn("DataFrame is being sorted!")

    x_timesteps = len(df.index)

    if x_timesteps - lookback - horizon - gap <= -1:
        raise ValueError("Not enough timesteps to build.")

    end = x_timesteps - horizon - gap + 1

    ids = np.arange(lookback, end)

    xframes = np.empty(shape=(len(ids) * lookback, df.shape[1]), dtype="object")
    xindex = [
        np.empty(shape=(len(ids) * lookback), dtype="<U31"),
        np.empty(shape=(len(ids) * lookback), dtype=df.index.dtype),
    ]

    yframes = np.empty(shape=(len(ids) * horizon, df.shape[1]), dtype="object")
    yindex = [
        np.empty(shape=(len(ids) * horizon), dtype="<U31"),
        np.empty(shape=(len(ids) * horizon), dtype=df.index.dtype),
    ]

    for i in ids:
        uuid_str = str(uuid.uuid4())

        # X
        frame = df.iloc[i - lookback : i]
        xframes[
            (i - lookback) * lookback : (i - lookback + 1) * lookback, :
        ] = frame.values
        # xindex[0][(i - lookback) * lookback : (i - lookback + 1) * lookback] = (
        #     i - lookback
        # ) * np.ones(lookback, dtype=int)
        xindex[0][
            (i - lookback) * lookback : (i - lookback + 1) * lookback
        ] = np.repeat(uuid_str, lookback)
        xindex[1][
            (i - lookback) * lookback : (i - lookback + 1) * lookback
        ] = frame.index.values

        # Y
        frame = df.iloc[i + gap : i + gap + horizon]
        yframes[
            (i - lookback) * horizon : (i - lookback + 1) * horizon, :
        ] = frame.values
        # yindex[0][(i - lookback) * horizon : (i - lookback + 1) * horizon] = (
        #     i - lookback
        # ) * np.ones(horizon, dtype=int)
        yindex[0][(i - lookback) * horizon : (i - lookback + 1) * horizon] = np.repeat(
            uuid_str, horizon
        )
        yindex[1][
            (i - lookback) * horizon : (i - lookback + 1) * horizon
        ] = frame.index.values

    timesteps_name = df.index.name or "timesteps"

    return Panel(
        xframes,
        columns=df.columns,
        index=pd.MultiIndex.from_arrays(xindex, names=["id", timesteps_name]),
    ), Panel(
        yframes,
        columns=df.columns,
        index=pd.MultiIndex.from_arrays(yindex, names=["id", timesteps_name]),
    )


# def reset_ids(panels: list[Panel], inplace: bool = False) -> list[Panel]:
#     """
#     Reset ids of a panel.

#     Args:
#         panels (``list``): List of panels
#         inplace (``bool``): Whether to reset ids inplace or not.

#     Returns:
#         ``Panel``: Reset id of panel
#     """

#     # Check if id in x and y are the same
#     # ! Broken
#     # if not all(np.array_equal(panels[0].ids, panel.ids) for panel in panels):
#     #     raise ValueError(
#     #         "Ids for panels are not the same. Try using match function first."
#     #     )

#     return [panel.reset_ids(inplace=inplace) for panel in panels]


def dropna_match(x, y):
    """
    Drop frames with NaN in panels and match ids.

    Args:
        x (``Panel``): Panel with x data
        y (``Panel``): Panel with y data

    Returns:
        ``Panel``: Panel with dropped frames and matched ids
    """

    x_t = x.dropna_frames()
    y_t = y.match_frames(x_t)

    y_t = y_t.dropna_frames()
    x_t = x_t.match_frames(y_t)

    return x_t, y_t


# ? Why is this function necessay if we can use pandas concat and append?
# def concat_panels(
#     panels: list[Panel], reset_ids: bool = False, sort: bool = False, axis=0
# ) -> Panel:
#     """
#     Concatenate panels.

#     Args:
#         panels (``list``): List of panels
#         reset_ids (``bool``): Whether to reset ids
#         sort (``bool``): Whether to sort by id
#         axis (``axis``): Same as pd.concat axis

#     Returns:
#         ``Panel``: Concatenated panels
#     """

#     # ! What if the panels have the same ids before concat and we want to concat anyway? IDs should be hashes.
#     # Get ids of all panels
#     # ids = list(chain(*[panel.ids for panel in panels]))

#     # Check duplicated ids in list
#     # if len(ids) != len(set(ids)):
#     #     raise ValueError("There are duplicated ids in the list.")

#     panel = Panel(pd.concat(panels, axis=axis))

#     if sort:
#         panel = panel.sort_panel()

#     if reset_ids:
#         panel.reset_ids(inplace=True)

#     return panel


def set_training_split(
    x: Panel,
    y: Panel,
    train_size: float | int = 0.7,
    val_size: float | int = 0.2,
    test_size: float | int = 0.1,
) -> None:
    """
    Splits panel in training, validation, and test.

    Args:
        train_size (``float`` or ``int``): Fraction of data to use for training.
        test_size (``float`` or ``int``): Fraction of data to use for testing.
        val_size (``float`` or ``int``): Fraction of data to use for validation.

    Example:

    >>> x, y = set_training_split(x, y, train_size=0.8, val_size=0.2, test_size=0.1)
    """

    if x.num_timesteps >= y.num_timesteps:
        x.set_training_split(
            train_size=train_size, val_size=val_size, test_size=test_size
        )
        y.train_size = x.train_size
        y.val_size = x.val_size
        y.test_size = x.test_size
    else:
        y.set_training_split(
            train_size=train_size, val_size=val_size, test_size=test_size
        )
        x.train_size = y.train_size
        x.val_size = y.val_size
        x.test_size = y.test_size


class _PanelSeries(pd.Series):
    def __init__(self, df, *args, **kwargs):
        super().__init__(df, *args, **kwargs)

    @property
    def _constructor_expanddim(self):
        # def f(*args, **kw):

        #     df = Panel(*args, **kw)

        #     # Workaround to fix pandas bug
        #     if (df.index.nlevels > 1 and self.index.nlevels > 1) and len(
        #         df.index.levels
        #     ) > len(self.index.levels):
        #         df = df.droplevel(0, axis="index")

        #     if df.num_frames == self.num_frames:
        #         self._copy_attrs(df)

        #     return df

        # return f

        return Panel

    @property
    def _constructor(self):
        return _PanelSeries


class Panel(pd.DataFrame):
    """
    Panel class.
    """

    def __init__(self, *args, **kw):
        super(Panel, self).__init__(*args, **kw)
        if len(args) == 1 and isinstance(args[0], Panel):
            args[0]._copy_attrs(self)

    _attributes_ = "train_size,test_size,val_size"

    def _copy_attrs(self, df):
        for attr in self._attributes_.split(","):
            df.__dict__[attr] = getattr(self, attr, None)

    @property
    def _constructor(self):
        def f(*args, **kw):

            # with contextlib.suppress(Exception):
            #     index = [a for a in args[0].axes if isinstance(a, pd.MultiIndex)]
            #     if index and len(index[0]) == self.num_timesteps:
            #         return pd.DataFrame(*args, **kw)

            df = Panel(*args, **kw)

            # Workaround to fix pandas bug
            if (df.index.nlevels > 1 and self.index.nlevels > 1) and len(
                df.index.levels
            ) > len(self.index.levels):
                df = df.droplevel(0, axis="index")

            if df.num_frames == self.num_frames:
                self._copy_attrs(df)

            # df.style.set_properties(**{"background-color": "black", "color": "green"})

            return df

        return f

    @property
    def _constructor_sliced(self):
        return _PanelSeries

    @property
    def num_frames(self) -> int:
        """Returns the number of frames in the panel."""
        return self.shape_panel[0]

    @property
    def num_timesteps(self) -> int:
        """Returns the number of timesteps in the panel."""
        return self.shape_panel[1]

    @property
    def num_columns(self) -> int:
        """Returns the number of columns in the panel."""
        return self.shape_panel[2]

    @property
    def frames(self) -> DataFrameGroupBy:
        """
        Returns panel's frames.
        """
        return self.groupby(level=0, as_index=True)

    @property
    def timesteps(self) -> pd.Int64Index:
        """
        Returns panel's timesteps.
        """
        return self.index.get_level_values(1)

    @property
    def ids(self) -> pd.Int64Index:
        """
        Returns panel's ids without duplicates.
        """
        return self.index.get_level_values(0).drop_duplicates()

    @ids.setter
    def ids(self, ids: list[int]) -> None:
        """
        Set panel's ids.

        Args:
            ids (``list``): List of ids.
        """

        ids = np.repeat(ids, self.shape_panel[1])
        timestep = self.index.get_level_values(1)

        index = pd.MultiIndex.from_arrays([ids, timestep], names=["id", timestep.name])

        self.index = index

    def reset_ids(self, inplace: bool = False) -> Panel | None:
        """
        Reset panel's ids.

        Args:
            inplace (``bool``): Whether to reset ids inplace.
        """
        new_ids = np.repeat(np.arange(self.num_frames), self.num_timesteps)
        new_index = pd.MultiIndex.from_arrays(
            [new_ids, self.index.get_level_values(1)],
            names=self.index.names,
        )

        return self.set_index(new_index, inplace=inplace)

    @property
    def shape_panel(self) -> tuple[int, int, int]:
        """
        Return a tuple representing the dimensionality of the Panel.
        """
        return (len(self.ids), int(self.shape[0] / len(self.ids)), self.shape[1])

    def row_panel(self, n: list[int] | int = 0) -> Panel:
        """
        Returns the nth row of each frame.

        Args:
            n (``list[int]`` or ``int``): Row index.
        """
        if isinstance(n, int):
            n = [n]

        if all(n < -1 or n >= self.num_timesteps for n in n):
            raise ValueError("n must be -1 or between 0 and the number of timesteps")

        new_panel = self.groupby(level=0, as_index=False).nth(n)
        self._copy_attrs(new_panel)
        return new_panel

    def get_timesteps(self, n: list[int] | int = 0) -> Panel:
        """
        Returns the first timestep of each frame in the panel.

        Args:
            n (``list[int]`` or ``int``): Timestep to return.
        """

        if isinstance(n, int):
            n = [n]

        if self.index.nlevels == 1:
            return self.frames.take(n).index.get_level_values(0)

        return self.frames.take(n).index.get_level_values(1)

    @property
    def values_panel(self) -> np.ndarray:
        """
        3D matrix with Panel value.

        Example:

        >>> panel.values
        array([[[283.95999146, 284.13000488, 280.1499939 , 281.77999878],
                [282.58999634, 290.88000488, 276.73001099, 289.98001099]],
               [[282.58999634, 290.88000488, 276.73001099, 289.98001099],
                [285.54000854, 286.3500061 , 274.33999634, 277.3500061 ]],
               [[285.54000854, 286.3500061 , 274.33999634, 277.3500061 ],
                [274.80999756, 279.25      , 271.26998901, 274.73001099]],
               [[274.80999756, 279.25      , 271.26998901, 274.73001099],
                [270.05999756, 272.35998535, 263.32000732, 264.57998657]]])
        """
        return np.reshape(self.to_numpy(), self.shape_panel)

    def flatten_panel(self) -> pd.DataFrame:
        """
        Flatten the panel.
        """

        new_timesteps = np.resize(
            np.arange(self.num_timesteps), self.num_timesteps * self.num_frames
        )
        new_index = pd.MultiIndex.from_arrays(
            [self.index.get_level_values(0), new_timesteps],
            names=self.index.names,
        )

        panel = (
            self.set_index(new_index)
            .reset_index()
            .pivot(index="id", columns=self.index.names[1])
        )

        columns = [f"{col}-{index}" for col, index in panel.columns.to_flat_index()]

        panel.columns = columns

        return panel

    def drop_ids(self, ids: list[int] | int, inplace: bool = False) -> Panel | None:
        """
        Drop frames by id.

        Args:
            ids (``list[int]`` or ``int``): List of ids to drop.
            inplace (``bool``): Whether to drop ids inplace.

        Returns:
            ``Panel``: Panel with frames dropped.
        """

        if self.index.nlevels == 1:
            return self.drop(index=ids, axis=0, inplace=inplace)

        return self.drop(index=ids, level=0, inplace=inplace)

    def findna_frames(self) -> pd.Int64Index:
        """
        Find NaN values index.

        Returns:
            ``List``: List with index of NaN frames.
        """
        na = self.isna().any(axis=1)
        return (
            self[na].index.get_level_values(0).drop_duplicates()
            if na.any()
            else pd.Int64Index([], name="id")
        )

    def dropna_frames(self, inplace: bool = False) -> Panel | None:
        """
        Drop frames with missing values from the panel.

        Args:
            inplace (``bool``): Whether to drop frames inplace.

        Returns:
            ``Panel``: Panel with frames dropped.
        """
        return self.drop_ids(self.findna_frames(), inplace=inplace)

    def match_frames(self, other: Panel, inplace: bool = False) -> Panel | None:
        """
        Match panel with other panel.

        This function will match the ids and id order of self based on the ids of other.

        Args:
            other (``Panel``): Panel to match with.
            inplace (``bool``): Whether to match inplace.

        Returns:
            ``Panel``: Result of match function.
        """

        other_ids = set(other.ids)
        self_ids = set(self.ids)

        if [i for i in other_ids if i not in self_ids]:
            raise ValueError("There are elements in other that are not in self.")

        if inplace:
            return self.drop_ids(self_ids - other_ids, inplace=True)

        return self.loc[other.ids]

    def set_training_split(
        self,
        train_size: float | int = 0.7,
        val_size: float | int = 0.2,
        test_size: float | int = 0.1,
    ) -> None:
        """
        Splits Panel into training, validation, and test.

        Args:
            train_size (``float`` or ``int``): Fraction of data to use for training.
            test_size (``float`` or ``int``): Fraction of data to use for testing.
            val_size (``float`` or ``int``): Fraction of data to use for validation.

        Example:

        >>> panel.set_training_split(train_size=0.8, val_size=0.2, test_size=0.1)
        """

        n_train, n_val, n_test = _validate_training_split(
            self.num_frames,
            train_size=train_size,
            val_size=val_size,
            test_size=test_size,
        )

        self.train_size = n_train
        self.val_size = n_val - self.num_timesteps + 1
        self.test_size = n_test - self.num_timesteps + 1

    def to_dataframe(self) -> pd.DataFrame:
        """
        Convert panel to dataframe.

        Returns:
            ``pd.DataFrame``: Dataframe with panel values.
        """
        return pd.DataFrame(self)

    @property
    def train(self) -> Panel:
        """
        Returns the Panel with the training set.

        Returns:
            ``Panel``: Panel with the training set.
        """

        return self[: self.train_size * self.num_timesteps] if self.train_size else None

    @train.setter
    def train(self, value: np.ndarray) -> None:
        """
        Set the training set.

        Args:
            value (``Panel``): Panel with the training set.
        """

        if not self.train_size:
            raise ValueError("No training set was set.")
        self[: self.train_size * self.num_timesteps] = value.values

    @property
    def val(self) -> Panel:
        """
        Returns the Panel with the validation set.

        Returns:
            ``Panel``: Panel with the validation set.
        """

        return (
            self[
                (self.train_size + self.num_timesteps - 1)
                * self.num_timesteps : (
                    self.train_size + self.val_size + self.num_timesteps - 1
                )
                * self.num_timesteps
            ]
            if self.val_size
            else None
        )

    @val.setter
    def val(self, value: np.ndarray) -> None:
        """
        Set the validation set.

        Args:
            value (``Panel``): Panel with the validation set.
        """

        if not self.val_size:
            raise ValueError("No validation set was set.")
        self[
            (self.train_size + self.num_timesteps - 1)
            * self.num_timesteps : (
                self.train_size + self.val_size + self.num_timesteps - 1
            )
            * self.num_timesteps
        ] = value.values

    @property
    def test(self) -> Panel:
        """
        Returns the Panel with the testing set.

        Returns:
            ``Panel``: Panel with the testing set.
        """

        return self[-self.test_size * self.num_timesteps :] if self.test_size else None

    @test.setter
    def test(self, value: np.ndarray) -> None:
        """
        Set the testing set.

        Args:
            value (``Panel``): Panel with the testing set.
        """

        if not self.test_size:
            raise ValueError("No testing set was set.")
        self[-self.test_size * self.num_timesteps :] = value.values

    def head_panel(self, n: int = 5) -> Panel:
        """
        Return the first n frames of the panel.

        Args:
            n (``int``): Number of frames to return.

        Returns:
            ``Panel``: Result of head function.
        """
        return self[: n * self.shape_panel[1]]

    def tail_panel(self, n: int = 5) -> Panel:
        """
        Return the last n frames of the panel.

        Args:
            n (``int``): Number of frames to return.

        Returns:
            ``Panel``: Result of tail function.
        """
        return self[-n * self.shape_panel[1] :]

    def sort_panel(
        self,
        ascending: bool = True,
        inplace: bool = False,
        kind: str = "quicksort",
        key: callable = None,
    ) -> Panel | None:
        """
        Sort panel by ids.

        Args:
            ascending (``bool`` or list-like of ``bools``, default True): Sort ascending vs. descending. When the index is a MultiIndex the sort direction can be controlled for each level individually.
            inplace (``bool``, default False): If True, perform operation in-place.
            kind ({'quicksort', 'mergesort', 'heapsort', 'stable'}, default 'quicksort'): Choice of sorting algorithm. See also numpy.sort() for more information. mergesort and stable are the only stable algorithms. For DataFrames, this option is only applied when sorting on a single column or label.
            key (callable, optional): If not None, apply the key function to the index values before sorting. This is similar to the key argument in the builtin sorted() function, with the notable difference that this key function should be vectorized. It should expect an Index and return an Index of the same shape. For MultiIndex inputs, the key is applied per level.

        Returns:
            ``Panel`` or ``None``: The original DataFrame sorted by the labels or None if `inplace=True`.
        """

        return self.sort_index(
            level=0,
            ascending=ascending,
            inplace=inplace,
            kind=kind,
            sort_remaining=False,
            key=key,
        )

    def sample_panel(
        self,
        samples: int | float = 5,
        how: str = "spaced",
        reset_ids: bool = False,
        seed: int = 42,
    ) -> Panel | None:
        """
        Sample panel returning a subset of frames.

        Args:
            samples (``int`` or ``float``): Number or percentage of samples to return.
            how (``str``): Sampling method, 'spaced' or 'random'
            reset_ids (``bool``): If True, reset the index of the sampled panel.
            seed (``int``): Random seed.

        Returns:
            ``Panel``: Result of sample function.
        """

        train_size = self.train_size if hasattr(self, "train_size") else self.num_frames
        val_size = self.val_size if hasattr(self, "val_size") else 0
        test_size = self.test_size if hasattr(self, "test_size") else 0

        train_samples, val_samples, test_samples = _validate_sample_panel(
            samples=samples,
            train_size=train_size,
            val_size=val_size,
            test_size=test_size,
        )

        if how == "random":
            # Set seed
            np.random.seed(seed)

            if hasattr(self, "train_size"):
                train_ids = sorted(
                    np.random.choice(self.train.ids, train_samples, replace=False)
                )
                val_ids = sorted(
                    np.random.choice(self.val.ids, val_samples, replace=False)
                )
                test_ids = sorted(
                    np.random.choice(self.test.ids, test_samples, replace=False)
                )
            else:
                train_ids = sorted(
                    np.random.choice(self.ids, train_samples, replace=False)
                )
                val_ids = []
                test_ids = []

        elif how == "spaced":
            if hasattr(self, "train_size"):
                train_ids = np.linspace(
                    self.train.ids[0],
                    self.train.ids[-1],
                    train_samples,
                    dtype=int,
                    endpoint=True,
                )
                val_ids = np.linspace(
                    self.val.ids[0],
                    self.val.ids[-1],
                    val_samples,
                    dtype=int,
                    endpoint=True,
                )
                test_ids = np.linspace(
                    self.test.ids[0],
                    self.test.ids[-1],
                    test_samples,
                    dtype=int,
                    endpoint=True,
                )
            else:
                train_ids = np.linspace(
                    self.ids[0], self.ids[-1], train_samples, dtype=int, endpoint=True
                )
                val_ids = []
                test_ids = []

        new_panel = self.loc[[*train_ids, *val_ids, *test_ids]]

        # Reset ids
        if reset_ids:
            new_panel.reset_ids(inplace=True)

        # Set new train, val, test sizes
        if hasattr(self, "train_size"):
            new_panel.train_size = train_samples
            new_panel.val_size = val_samples
            new_panel.test_size = test_samples

        # # TODO inplace not working
        # if inplace:
        #     self = new_panel
        #     return None

        return new_panel

    def shuffle_panel(self, seed: int = None, reset_ids: bool = False) -> Panel | None:
        """
        Shuffle the panel.

        Args:
            seed (``int``): Random seed.
            reset_ids (``bool``): If True, reset the index of the shuffled panel.

        Returns:
            ``Panel``: Result of shuffle function.
        """

        # warnings.warn("Shuffling the panel can result in data leakage.")

        if hasattr(self, "train_size"):
            train_ids = list(self.train.ids)
            val_ids = list(self.val.ids)
            test_ids = list(self.test.ids)
        else:
            train_ids = list(self.ids)
            val_ids = []
            test_ids = []

        random.seed(seed)
        random.shuffle(train_ids)
        random.shuffle(val_ids)
        random.shuffle(test_ids)

        new_panel = self.loc[[*train_ids, *val_ids, *test_ids]]

        # Reset ids
        if reset_ids:
            new_panel.reset_ids(inplace=True)

        # # TODO inplace not working
        # if inplace:
        #     self = new_panel
        #     return None

        return new_panel

    # ! Inconsistent with lookback >= 2 if frames were modified.
    # def plot(
    #     self,
    #     add_annotation: bool = True,
    #     max: int = 10_000,
    #     use_timestep: bool = False,
    #     **kwargs,
    # ) -> plot.PanelFigure:
    #     """
    #     Plot the panel.

    #     Args:
    #         add_annotation (``bool``): If True, plot the training, validation, and test annotation.
    #         max (``int``): Maximum number of samples to plot.
    #         use_timestep (``bool``): If True, plot the timestep instead of the sample index.
    #         **kwargs: Additional arguments to pass to the plot function.

    #     Returns:
    #         ``plot``: Result of plot function.
    #     """

    #     panel = self.row_panel(n=0)

    #     if max and self.num_frames > max:
    #         return plot(
    #             panel.sample_panel(max, how="spaced"),
    #             use_timestep=use_timestep,
    #             add_annotation=add_annotation,
    #             **kwargs,
    #         )
    #     return plot(
    #         panel, use_timestep=use_timestep, add_annotation=add_annotation, **kwargs
    #     )



from __future__ import annotations

import warnings

import numpy as np
import pandas as pd
import tensorflow as tf
from keras import Sequential
from keras.layers import Conv1D, Dense, Flatten, Reshape
from sklearn.base import is_classifier
from sklearn.metrics import auc, roc_curve

from .panel import Panel, set_training_split


class _ConstantKerasModel(tf.keras.Model):
    """A Keras model that returns the input values as outputs."""

    def __init__(self):
        super().__init__()

    def call(self, inputs):
        return inputs


class BaseModel:
    """Base class for panel models."""

    def __init__(
        self,
        x: Panel,
        y: Panel,
        model_type: str = None,
        loss: str = None,
        optimizer: str = None,
        metrics: list[str] = None,
        last_activation: str = None,
    ):
        """
        Base model class.

        Args:
            x (``Panel``): Panel of input data.
            y (``Panel``): Panel of output data.
            model_type (``str``): Type of model.
            loss (``str``): Loss function.
            optimizer (``str``): Optimizer.
            metrics (``list[str]``): Metrics.
            last_activation (``str``): Last activation.
        """

        PARAMS = {
            "regression": {
                "loss": "MSE",
                "optimizer": "adam",
                "metrics": ["mae"],
                "last_activation": "linear",
            },
            "classification": {
                "loss": "binary_crossentropy",
                "optimizer": "adam",
                "metrics": ["AUC", "accuracy"],
                "last_activation": "sigmoid",
            },
            "multi_classification": {
                "loss": "categorical_crossentropy",
                "optimizer": "adam",
                "metrics": ["AUC", "accuracy"],
                "last_activation": "softmax",
            },
        }

        # Raise error when panel has nan values
        if x.findna_frames().any():
            raise ValueError("Panel x has NaN values.")
        if y.findna_frames().any():
            raise ValueError("Panel y has NaN values.")

        # Convert boolean in x and y to int
        for col in x.columns:
            if x[col].dtype == bool:
                x[col] = x[col].astype(int)
        for col in y.columns:
            if y[col].dtype == bool:
                y[col] = y[col].astype(int)

        # Get model_type automatically from y cardinality
        if y.shape[1] == 1:
            if not model_type and y.iloc[:, 0].unique().shape[0] == 2:
                model_type = "classification"
            elif not model_type and y.iloc[:, 0].unique().shape[0] < 20:
                model_type = "multi_classification"
            elif not model_type:
                model_type = "regression"
        if not model_type:
            raise ValueError("Model type not specified.")

        # Run set_training_split on x and y if not already done
        if (
            not hasattr(x, "train_size")
            or not x.train_size
            or not hasattr(y, "train_size")
            or not y.train_size
        ):
            warnings.warn("Running set_training_split with default parameters")

            set_training_split(x, y)

        # Raise error if column is not numeric
        for sample in [x, y]:
            for col in sample.columns:
                if sample[col].dtype not in [np.float64, np.int64]:
                    raise ValueError(
                        f"Column {col} is of type {sample[col].dtype}, not numeric."
                    )

        self.x = x
        self.y = y

        self.model_type = model_type
        self.loss = loss or PARAMS[model_type]["loss"]
        self.optimizer = optimizer or PARAMS[model_type]["optimizer"]
        self.metrics = metrics or PARAMS[model_type]["metrics"]
        self.last_activation = last_activation or PARAMS[model_type]["last_activation"]

        self.set_arrays()
        self.build()
        self.compile()
        self.model._name = self.__class__.__name__

    def set_arrays(self) -> None:
        """Set the arrays."""
        self.x_train = self.x.train.values_panel
        self.x_val = self.x.val.values_panel
        self.x_test = self.x.test.values_panel

        self.y_train = self.y.train.values_panel.squeeze().reshape(
            self.x_train.shape[0], -1
        )
        self.y_val = self.y.val.values_panel.squeeze().reshape(self.x_val.shape[0], -1)
        self.y_test = self.y.test.values_panel.squeeze().reshape(
            self.x_test.shape[0], -1
        )

    def get_auc(self) -> float:
        """Get the AUC score."""
        y = self.y_test.squeeze()
        prediction = self.model.predict(self.x_test).squeeze()
        fpr, tpr, _ = roc_curve(y, prediction)
        fpr, tpr, _ = roc_curve(y, prediction)
        return auc(fpr, tpr)

    def fit(self, **kwargs) -> None:
        """
        Fit the model.

        Args:
            **kwargs: Additional arguments to pass to the fit method.
        """
        self.model.fit(
            self.x_train,
            self.y_train,
            validation_data=(self.x_val, self.y_val),
            **kwargs,
        )

    def compile(self, **kwargs) -> None:
        """Compile the model.

        Args:
            **kwargs: Additional arguments to pass to the compile method.
        """
        self.model.compile(
            loss=self.loss, optimizer=self.optimizer, metrics=self.metrics, **kwargs
        )

    def build(self) -> None:
        """Build the model."""
        pass

    def predict_proba(self, data: Panel = None, **kwargs) -> Panel:
        """Predict probabilities.

        Args:
            data (``Panel``): Panel of data to predict.
            **kwargs: Additional arguments to pass to the predict method.

        Returns:
            Panel of predicted probabilities.
        """

        if data is not None:
            x = data.values_panel
            if data.index.nlevels == 1:
                index = pd.Index(data.get_timesteps(0))
            else:
                index = pd.MultiIndex.from_arrays([data.ids, data.get_timesteps(0)])
        else:
            x = np.concatenate([self.x_train, self.x_val, self.x_test], axis=0)
            if self.y.index.nlevels == 1:
                index = pd.Index(
                    np.concatenate(
                        [self.y.train.index, self.y.val.index, self.y.test.index]
                    ),
                    name=self.y.index.name,
                )
            else:
                index = pd.MultiIndex.from_tuples(
                    np.concatenate(
                        [self.y.train.index, self.y.val.index, self.y.test.index]
                    ),
                    names=self.y.index.names,
                )

        return Panel(
            self.model.predict(x),
            columns=self.y.columns,
            index=index,
        )

    def predict(self, data: Panel = None, **kwargs) -> Panel:
        """Predict.

        Args:
            data (``Panel``): Panel of data to predict.
            **kwargs: Additional arguments to pass to the predict method.

        Returns:
            ``Panel`` of predicted values.
        """

        threshold = self.get_auc() if self.model_type == "classification" else None

        panel = self.predict_proba(data=data, **kwargs)

        return (
            panel if threshold is None else panel.apply(lambda x: (x > threshold) + 0)
        )

    def score(self, on: list[str] | str = None, **kwargs) -> pd.DataFrame:
        """Score the model.

        Args:
            on (``list[str]`` or ``str``): Columns to score on.
            **kwargs: Additional arguments to pass to the score method.

        Returns:
            Panel of scores.
        """
        on = [on] if on else ["train", "val", "test"]

        dic = {}
        if "train" in on:
            dic["train"] = self.model.evaluate(
                self.x_train, self.y_train, verbose=0, **kwargs
            )
        if "test" in on:
            dic["test"] = self.model.evaluate(
                self.x_test, self.y_test, verbose=0, **kwargs
            )
        if "val" in on:
            dic["val"] = self.model.evaluate(
                self.x_val, self.y_val, verbose=0, **kwargs
            )

        indexes = [
            self.model.metrics_names.index(metric.lower()) for metric in self.metrics
        ]

        return pd.DataFrame(
            {key: [value[index] for index in indexes] for key, value in dic.items()},
            index=self.metrics,
        )

    def residuals(self) -> Panel:
        """Residuals.

        Returns:
            ``Panel`` of residuals.
        """
        return self.predict() - self.y


class _Baseline(BaseModel):
    def __init__(
        self,
        x,
        y,
        model_type: str = None,
        loss: str = None,
        metrics: list[str] = None,
    ):

        super().__init__(x=x, y=y, model_type=model_type, loss=loss, metrics=metrics)

    def build(self) -> None:
        """Build the model."""
        self.model = _ConstantKerasModel()


class BaselineShift(_Baseline):
    """Baseline shift model."""

    # ! Maybe shift should be y.horizon by default, to avoid leakage
    # TODO test with different gap and horizon values

    def __init__(
        self,
        x,
        y,
        model_type: str = None,
        loss: str = None,
        metrics: list[str] = None,
        fillna=0,
        shift=1,
    ):

        self.fillna = fillna
        self.shift = shift
        super().__init__(x=x, y=y, model_type=model_type, loss=loss, metrics=metrics)

    def set_arrays(self):
        """Set the arrays."""
        self.x_train = (
            self.y.train.shift(self.shift * self.x.num_timesteps)
            .fillna(self.fillna)
            .values
        )
        self.x_val = (
            self.y.val.shift(self.shift * self.x.num_timesteps)
            .fillna(self.fillna)
            .values
        )
        self.x_test = (
            self.y.test.shift(self.shift * self.x.num_timesteps)
            .fillna(self.fillna)
            .values
        )

        self.y_train = self.y.train.values
        self.y_val = self.y.val.values
        self.y_test = self.y.test.values

    def build(self) -> None:
        """Build the model."""
        self.model = _ConstantKerasModel()


class BaselineConstant(_Baseline):
    """Baseline constant model."""

    # TODO BUG: Not working when model_type="classification"
    def __init__(
        self,
        x,
        y,
        model_type: str = None,
        loss: str = None,
        metrics: list[str] = None,
        constant: float = 0,
    ):

        self.constant = constant if model_type == "regression" else int(constant)
        super().__init__(x=x, y=y, model_type=model_type, loss=loss, metrics=metrics)

    def set_arrays(self) -> None:
        """Set the arrays."""
        self.x_train = np.full(self.y.train.shape, self.constant)
        self.x_val = np.full(self.y.val.shape, self.constant)
        self.x_test = np.full(self.y.test.shape, self.constant)

        self.y_train = self.y.train.values
        self.y_val = self.y.val.values
        self.y_test = self.y.test.values


class DenseModel(BaseModel):
    """Dense model."""

    def __init__(
        self,
        x,
        y,
        model_type: str = None,
        dense_layers: int = 1,
        dense_units: int = 32,
        activation: str = "relu",
        loss: str = None,
        optimizer: str = None,
        metrics: list[str] = None,
        last_activation: str = None,
    ):
        """
        Dense Model.

        Args:
            panel (``Panel``): Panel with data
            model_type (``str``): Model type (regression, classification, multi_classification)
            dense_layers (``int``): Number of dense layers
            dense_units (``int``)t``)t``)t``)t``): Number of neurons in each dense layer
            activation (``str``): Activation type of each dense layer
            loss (``str``): Loss name
            optimizer (``str``): Optimizer name
            metrics (``list[str]``): Metrics list
            last_activation (``str``): Activation type of the last layer

        Returns:
            ``DenseModel``: Constructed DenseModel
        """

        self.dense_layers = dense_layers
        self.dense_units = dense_units
        self.activation = activation

        super().__init__(
            x=x,
            y=y,
            model_type=model_type,
            loss=loss,
            optimizer=optimizer,
            metrics=metrics,
            last_activation=last_activation,
        )

    def build(self) -> None:
        """Build the model."""
        dense = Dense(units=self.dense_units, activation=self.activation)
        layers = [Flatten()]  # (time, features) => (time*features)
        layers += [dense for _ in range(self.dense_layers)]
        # layers += [drop]
        layers += [
            Dense(
                units=self.y.num_timesteps * self.y.num_columns,
                activation=self.last_activation,
            ),
            Reshape((self.y.num_columns,)),
        ]

        self.model = Sequential(layers)


class ConvModel(BaseModel):
    """Convolutional model."""

    def __init__(
        self,
        x: Panel,
        y: Panel,
        model_type: str = None,
        conv_layers: int = 1,
        conv_filters: int = 32,
        kernel_size: int = 3,
        dense_layers: int = 1,
        dense_units: int = 32,
        activation: str = "relu",
        loss: str = None,
        optimizer: str = None,
        metrics: list[str] = None,
        last_activation: str = None,
    ):
        """
        Convolution Model.

        Args:
            x (``Panel``): Panel with x data
            y (``Panel``): Panel with y data
            model_type (``str``): Model type (regression, classification, multi_classification)
            conv_layers (``int``): Number of convolution layers
            conv_filters (``int``): Number of convolution filters
            kernel_size (``int``): Kernel size of convolution layer
            dense_layers (``int``): Number of dense layers
            dense_units (``int``): Number of neurons in each dense layer
            activation (``str``): Activation type of each dense layer
            loss (``str``): Loss name
            optimizer (``str``): Optimizer name
            metrics (``list[str]``): Metrics list
            last_activation (``str``): Activation type of the last layer

        Returns:
            ``DenseModel``: Constructed DenseModel
        """

        if x.shape_panel[1] < kernel_size:
            raise ValueError(
                f"Lookback ({x.shape_panel[1]}) must be greater or equal to kernel_size ({kernel_size})"
            )

        self.conv_layers = conv_layers
        self.conv_filters = conv_filters
        self.kernel_size = kernel_size
        self.dense_layers = dense_layers
        self.dense_units = dense_units
        self.activation = activation

        super().__init__(
            x=x,
            y=y,
            model_type=model_type,
            loss=loss,
            optimizer=optimizer,
            metrics=metrics,
            last_activation=last_activation,
        )

    def build(self) -> None:
        """Build the model."""

        if self.x.num_timesteps % self.kernel_size != 0:
            warnings.warn("Kernel size is not a divisor of lookback.")

        conv = Conv1D(
            filters=self.conv_filters,
            kernel_size=self.kernel_size,
            activation=self.activation,
        )

        dense = Dense(units=self.dense_units, activation=self.activation)

        layers = [conv for _ in range(self.conv_layers)]
        layers += [Flatten()]
        layers += [conv for _ in range(self.conv_layers)]
        layers += [dense for _ in range(self.dense_layers)]
        layers += [
            Dense(
                units=self.y.num_timesteps * self.y.num_columns,
                activation=self.last_activation,
            ),
            Reshape((self.y.num_columns,)),
        ]

        self.model = Sequential(layers)


class LinearRegression(DenseModel):
    """Linear regression model."""

    def __init__(self, x, y, **kwargs):
        super().__init__(x=x, y=y, model_type="regression", dense_layers=0, **kwargs)


class LogisticRegression(DenseModel):
    """Logistic regression model."""

    def __init__(self, x, y, **kwargs):
        super().__init__(
            x=x, y=y, model_type="classification", dense_layers=0, **kwargs
        )


class ShallowModel:
    """Shallow model."""

    def __init__(self, x: Panel, y: Panel, model: str, metrics: list[str], **kwargs):
        """Shallow Model.

        Args:
            x (``Panel``): Panel with x data
            y (``Panel``): Panel with y data
            model (``str``): Model (regression, classification, multi_classification)
            metrics (``list[str]``): Metrics list
            **kwargs: Additional arguments

        Returns:
            ``ShallowModel``: Constructed ShallowModel
        """

        self.x = x
        self.y = y

        if len(self.y.columns) > 1:
            raise ValueError("ShallowModel can only be used for single-output models.")

        self.model = model(**kwargs)
        self.metrics = metrics
        self.set_arrays()

    def set_arrays(self) -> None:
        """
        Sets arrays for training, testing, and validation.
        """

        self.x_train = self.x.train.flatten_panel().values
        self.y_train = self.y.train.values

        self.x_val = self.x.val.flatten_panel().values
        self.y_val = self.y.val.values

        self.x_test = self.x.test.flatten_panel().values
        self.y_test = self.y.test.values

    def fit(self, **kwargs):
        """Fit the model.

        Args:
            **kwargs: Keyword arguments for the fit method of the model.
        """
        self.model.fit(X=self.x_train, y=self.y_train, **kwargs)

    def get_auc(self) -> float:
        """Get the AUC score."""

        y = self.y_test.squeeze()
        prediction = self.model.predict(self.x_test).squeeze()
        fpr, tpr, _ = roc_curve(y, prediction)
        return auc(fpr, tpr)

    def predict_proba(self, data: Panel = None) -> Panel:
        """Predict probabilities.

        Args:
            data (``Panel``): Panel with data

        Returns:
            ``ShallowModel``: The predicted probabilities.
        """

        if data is not None:
            x = data.flatten_panel().values
            if data.index.nlevels == 1:
                index = pd.Index(data.get_timesteps(0))
            else:
                index = pd.MultiIndex.from_arrays([data.ids, data.get_timesteps(0)])
        else:
            x = np.concatenate([self.x_train, self.x_val, self.x_test], axis=0)
            if self.y.index.nlevels == 1:
                index = pd.Index(
                    np.concatenate(
                        [self.y.train.index, self.y.val.index, self.y.test.index]
                    ),
                    name=self.y.index.name,
                )
            else:
                index = pd.MultiIndex.from_tuples(
                    np.concatenate(
                        [self.y.train.index, self.y.val.index, self.y.test.index]
                    ),
                    names=self.y.index.names,
                )

        output = self.model.predict_proba(x)

        if output.shape[1] == 2:
            output = output[:, 1]

            return Panel(
                output,
                columns=self.y.columns,
                index=index,
            )

        return Panel(
            output,
            columns=[f"{i}_prob" for i in range(output.shape[1])],
            index=index,
        )

    def predict(self, data: Panel = None) -> Panel:
        """Predict on data.

        Args:
            data (``Panel``, optional): Data to predict on. Defaults to None.

        Returns:
            ``Panel``: Predicted data
        """
        if is_classifier(self.model):
            threshold = self.get_auc()
            panel = self.predict_proba(data)
            return panel.apply(lambda x: (x > threshold) + 0)

        else:
            if data is not None:
                x = data.flatten_panel().values
                index = pd.MultiIndex.from_arrays([data.ids, data.first_timestamp])
            else:
                x = np.concatenate([self.x_train, self.x_val, self.x_test], axis=0)
                index = pd.MultiIndex.from_tuples(
                    np.concatenate(
                        [self.y.train.index, self.y.val.index, self.y.test.index]
                    ),
                    names=self.y.index.names,
                )

            return Panel(
                self.model.predict(x),
                columns=self.y.columns,
                index=index,
            )

    def score(self, on: list[str] | str = None) -> pd.DataFrame:
        """Score the model.

        Args:
            on (``list[str]`` or ``str``): Data to use for scoring

        Returns:
            ``pd.Series``: Score
        """
        on = [on] if on else ["train", "val", "test"]

        dic = {}

        if "train" in on:
            metrics_dict = {
                a.__name__: a(
                    self.y.train.values.squeeze(),
                    self.predict(self.x.train).values.squeeze(),
                )
                for a in self.metrics
            }

            dic["train"] = metrics_dict
        if "test" in on:
            metrics_dict = {
                a.__name__: a(
                    self.y.test.values.squeeze(),
                    self.predict(self.x.test).values.squeeze(),
                )
                for a in self.metrics
            }

            dic["test"] = metrics_dict
        if "val" in on:
            metrics_dict = {
                a.__name__: a(
                    self.y.val.values.squeeze(),
                    self.predict(self.x.val).values.squeeze(),
                )
                for a in self.metrics
            }

            dic["val"] = metrics_dict

        return pd.DataFrame(dic, index=[a.__name__ for a in self.metrics])

    def residuals(self) -> Panel:
        """Residuals.

        Returns:
            ``Panel``: Residuals
        """
        return self.predict() - self.y


def compute_score_per_model(*models, on="val"):
    # BUG
    """
    Compute score per model

    Args:
        *models: Models to score
        on (``str``, optional): Data to use for scoring. Defaults to "val".

    Returns:
        pd.DataFrame: Scores
    """

    return pd.DataFrame(
        [model.score(on=on) for model in models],
        index=[model.model.name for model in models],
    )


def compute_default_scores(x, y, model_type, epochs=10, verbose=0, **kwargs):
    # BUG
    """
    Compute default scores for a model.

    Args:
        x (``Panel``): X data.
        y (``Panel``): Y data.
        model_type (``str``): Model type.
        epochs (``int``, optional): Number of epochs. Defaults to 10.
        verbose (``int``, optional): Verbosity. Defaults to 0.
        **kwargs: Keyword arguments for the model.

    Returns:
        pd.DataFrame: Scores
    """
    models = [BaselineConstant, BaselineShift, DenseModel]
    models = [model(x=x, y=y, model_type=model_type) for model in models]
    for model in models:
        model.fit(epochs=epochs, verbose=verbose, **kwargs)
    return compute_score_per_model(*models)



from __future__ import annotations

import numpy as np


def _validate_training_split(
    n_samples: int,
    train_size: int | float,
    val_size: int | float,
    test_size: int | float,
) -> tuple[int, int, int]:
    """
    Validation helper to check if the split sizes are meaningful to the
    size of the data (n_samples)

    Args:
        n_samples (``int``): Number of samples in the data
        train_size (``int``): Size of the training set
        val_size (``int``): Size of the validation set
        test_size (``int``): Size of the test set

    Returns:
        Tuple[int, int, int]: Tuple of the training, validation, and test sizes
    """

    if sum(x is None for x in (train_size, val_size, test_size)) > 1:
        raise ValueError(
            "At least one from train_size, val_size and test_size can be None"
        )

    train_size_type = np.asarray(train_size).dtype.kind
    test_size_type = np.asarray(test_size).dtype.kind
    val_size_type = np.asarray(val_size).dtype.kind

    # Check train_size
    if (
        train_size_type == "i"
        and (train_size >= n_samples or train_size <= 0)
        or train_size_type == "f"
        and (train_size <= 0 or train_size >= 1)
    ):
        raise ValueError(
            "train_size={0} should be either positive and smaller"
            " than the number of samples {1} or a float in the "
            "(0, 1) range".format(train_size, n_samples)
        )

    # Check val_size
    if (
        val_size_type == "i"
        and (val_size >= n_samples or val_size <= 0)
        or val_size_type == "f"
        and (val_size <= 0 or val_size >= 1)
    ):
        raise ValueError(
            "val_size={0} should be either positive and smaller"
            " than the number of samples {1} or a float in the "
            "(0, 1) range".format(val_size, n_samples)
        )

    # Check test_size
    if (
        test_size_type == "i"
        and (test_size >= n_samples or test_size <= 0)
        or test_size_type == "f"
        and (test_size <= 0 or test_size >= 1)
    ):
        raise ValueError(
            "test_size={0} should be either positive and smaller"
            " than the number of samples {1} or a float in the "
            "(0, 1) range".format(test_size, n_samples)
        )

    if train_size is not None and train_size_type not in ("i", "f"):
        raise ValueError(f"Invalid value for train_size: {train_size}")
    if test_size is not None and test_size_type not in ("i", "f"):
        raise ValueError(f"Invalid value for test_size: {test_size}")
    if val_size is not None and val_size_type not in ("i", "f"):
        raise ValueError(f"Invalid value for val_size: {val_size}")

    # Check if percentages sum to less than 1
    if (
        train_size_type == "f"
        and val_size_type == "f"
        and test_size_type == "f"
        and train_size + val_size + test_size > 1
    ):
        raise ValueError(
            f"""The sum of train_size, val_size and test_size =
            {train_size + val_size + test_size}, should be in the (0, 1) range.
            Reduce train_size and/or val_size and/or test_size."""
        )

    if train_size_type == "f":
        n_train = round(train_size * n_samples)
    elif train_size_type == "i":
        n_train = float(train_size)

    if val_size_type == "f":
        n_val = round(val_size * n_samples)
    elif val_size_type == "i":
        n_val = float(val_size)

    if test_size_type == "f":
        n_test = round(test_size * n_samples)
    elif test_size_type == "i":
        n_test = float(test_size)

    # Calculate the number of samples in each split if any is None
    if not train_size:
        n_train = max(n_samples - n_val - n_test, 0)
    elif not val_size:
        n_val = max(n_samples - n_train - n_test, 0)
    elif not test_size:
        n_test = max(n_samples - n_train - n_val, 0)

    if abs(n_train + n_val + n_test - n_samples) == 1:
        n_test = max(n_samples - n_train - n_val, 0)

    if n_train + n_val + n_test != n_samples:
        raise ValueError(
            f"""The sum of train_size, val_size and test_size = {n_train + n_val + n_test},
            should be of the same size of samples {n_samples}.
            Change train_size and/or val_size and/or test_size."""
        )

    return int(n_train), int(n_val), int(n_test)


def _validate_sample_panel(
    samples: int, train_size: int | float, val_size: int | float, test_size: int | float
) -> tuple[int, int, int]:
    """
    Validation helper to check if the samples size is meaningful wrt to the
    size of the data (n_samples)

    Args:
        samples (``int``): Number of samples in the data
        train_size (``int``): Size of the training set
        val_size (``int``): Size of the validation set

    Returns:
        Tuple[int, int, int]: Tuple of the training, validation, and test sizes
    """

    n_samples = train_size + val_size + test_size
    samples_type = np.asarray(samples).dtype.kind

    # Check samples size
    if (
        samples_type == "i"
        and (samples >= n_samples or samples <= 0)
        or samples_type == "f"
        and (samples <= 0 or samples >= 1)
    ):
        raise ValueError(
            "samples={0} should be either positive and smaller"
            " than the number of samples {1} or a float in the "
            "(0, 1) range".format(samples, n_samples)
        )

    if samples is not None and samples_type not in ("i", "f"):
        raise ValueError(f"Invalid value for samples: {samples}")

    if samples_type == "i":
        samples /= n_samples

    train_samples = round(samples * train_size)
    val_samples = round(samples * val_size)
    test_samples = round(samples * test_size)

    return train_samples, val_samples, test_samples



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['wavy']

package_data = \
{'': ['*']}

install_requires = \
['ipykernel>=6.15.1,<7.0.0',
 'nbformat>=5.4.0,<6.0.0',
 'numpy>=1.22.3,<2.0.0',
 'pandas==1.4.3',
 'plotlab>=0.1.0,<0.2.0',
 'plotly>=5.7.0,<6.0.0',
 'sklearn>=0.0,<0.1',
 'tensorflow>=2.8.0,<3.0.0',
 'toml>=0.10.2,<0.11.0',
 'tqdm>=4.64.0,<5.0.0']

setup_kwargs = {
    'name': 'wavyts',
    'version': '0.1.10',
    'description': 'Wavy is a library to facilitate time series analysis',
    'long_description': '<!-- Title -->\n## ≈ Wavy: Time-Series Manipulation ≈\n\n<p>\n<img alt="GitHub Contributors" src="https://img.shields.io/github/contributors/logspace-ai/wavy" />\n<img alt="GitHub Last Commit" src="https://img.shields.io/github/last-commit/logspace-ai/wavy" />\n<!-- <img alt="GitHub Language Count" src="https://img.shields.io/github/languages/count/logspace-ai/wavy" /> -->\n<img alt="" src="https://img.shields.io/github/repo-size/logspace-ai/wavy" />\n<!-- <img alt="GitHub Issues" src="https://img.shields.io/github/issues/logspace-ai/wavy" /> -->\n<!-- <img alt="GitHub Closed Issues" src="https://img.shields.io/github/issues-closed/logspace-ai/wavy" /> -->\n<!-- <img alt="GitHub Pull Requests" src="https://img.shields.io/github/issues-pr/logspace-ai/wavy" /> -->\n<!-- <img alt="GitHub Closed Pull Requests" src="https://img.shields.io/github/issues-pr-closed/logspace-ai/wavy" />  -->\n<!-- <img alt="GitHub Commit Activity (Year)" src="https://img.shields.io/github/commit-activity/y/logspace-ai/wavy" /> -->\n<img alt="Github License" src="https://img.shields.io/github/license/logspace-ai/wavy" />  \n</p>\n\n\nWavy is a time-series manipulation library designed to simplify the pre-processing steps and reliably avoid the problem of data leakage. Its main structure is built on top of Pandas. <a href="https://logspace-ai.github.io/wavy/"><strong>Explore the docs 📖</strong></a>\n    <a href="https://github.com/logspace-ai/wavy">\n        <img width="50%" src="https://github.com/logspace-ai/wavy/blob/main/images/logo.png?raw=true" alt="Logo" width="419" height="235" align="right"></a>\n\n  \n\n<!-- PROJECT LOGO -->\n<!-- <div align="right">\n  <a href="https://github.com/logspace-ai/wavy">\n    <img width="49%" src="https://github.com/logspace-ai/wavy/blob/main/images/logo.png?raw=true" alt="Logo" width="419" height="235">\n  </a>\n\n</div> -->\n\n<!-- GETTING STARTED -->\n## 📦 Installation\n\nYou can install Wavy from pip:\n\n```bash\npip install wavyts\n```\n\n<!-- GETTING STARTED -->\n## 🚀 Quickstart\n\n```python\nimport numpy as np\nimport pandas as pd\nimport wavy\nfrom wavy import models\n\n# Start with any time-series dataframe:\ndf = pd.DataFrame({\'price\': np.random.randn(1000)}, index=range(1000))\n\n# Create panels. Each panel is a frame collection.\nx, y = wavy.create_panels(df, lookback=3, horizon=1)\n\n# x and y contain the past and corresponding future data.\n# lookback and horizon are the number of timesteps.\nprint("Lookback:", x.num_timesteps)\nprint("Horizon:", y.num_timesteps)\n\n# Set train-val-test split. Defaults to 0.7, 0.2 and 0.1, respectively.\nwavy.set_training_split(x, y)\n\n# Instantiate a model:\nmodel = models.LinearRegression(x, y)\nmodel.score()\n```\n    \n\n<!-- Description -->\n## Features\n\n💡 Wavy **is**:\n\n- A **resourceful**, **high-level** package with tools for time-series processing, visualization, and modeling.\n- A facilitator for **time-series windowing** that helps reduce boilerplate code and avoid shape confusion.\n\n❗ Wavy **is not**:\n\n- An efficient, performance-first framework (**yet!**).\n- Primarily focused on models. Processed data can be easily converted to numpy arrays for further exploration.\n\n\n<!-- CONTRIBUTING -->\n## Contributing\n\nContributions are what make the open source community such an amazing place to learn, inspire, and create. Any contributions you make to `wavy` are **greatly appreciated**.\n\nIf you have a suggestion that would make this better, please fork the repo and create a pull request. You can also simply open an issue with the tag "enhancement".\nDon\'t forget to give the project a star! ⭐\n\n1. Fork the Project\n2. Create your Feature Branch (`git checkout -b feature/AmazingFeature`)\n3. Commit your Changes (`git commit -m \'Add some AmazingFeature\'`)\n4. Push to the Branch (`git push origin feature/AmazingFeature`)\n5. Open a Pull Request\n\n\n<!-- LICENSE -->\n## License\n\nDistributed under the MIT License. See `LICENSE.txt` for more information.\n\n\n<!-- MARKDOWN LINKS & IMAGES -->\n<!-- https://www.markdownguide.org/basic-syntax/#reference-style-links -->\n[contributors-shield]: https://img.shields.io/github/contributors/logspace-ai/wavy.svg?style=for-the-badge\n[contributors-url]: https://github.com/logspace-ai/wavy/graphs/contributors\n[forks-shield]: https://img.shields.io/github/forks/logspace-ai/wavy.svg?style=for-the-badge\n[forks-url]: https://github.com/logspace-ai/wavy/network/members\n[stars-shield]: https://img.shields.io/github/stars/logspace-ai/wavy.svg?style=for-the-badge\n[stars-url]: https://github.com/logspace-ai/wavy/stargazers\n[issues-shield]: https://img.shields.io/github/issues/logspace-ai/wavy.svg?style=for-the-badge\n[issues-url]: https://github.com/logspace-ai/wavy/issues\n[license-shield]: https://img.shields.io/github/license/logspace-ai/wavy.svg?style=for-the-badge\n[license-url]: https://github.com/logspace-ai/wavy/blob/main/LICENSE.txt\n<!-- [documentation-url]: https://logspace-ai.github.io/wavy/ -->\n',
    'author': 'Ibis Prevedello',
    'author_email': 'ibiscp@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/logspace-ai/wavy',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
